# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Guoye Yang <498731903@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import os
import numpy as np
from jittor import compile_extern
# TODO: compare with pytorch

from jittor.test.test_log import find_log_with_re
if jt.has_cuda:
    from jittor.compile_extern import cublas_ops, cudnn_ops
else:
    cublas_ops = cudnn_ops = None


def conv_nchw(x, in_planes, out_planes, kernel_size, padding, stride=1, dilation=1, groups=1, init_method=None, w_=None):
    N,C,H,W = x.shape
    Kh, Kw = kernel_size, kernel_size
    G = groups
    CpG = C // G # channels per group
    padding = (padding, padding)
    dilation = (dilation, dilation)
    stride = (stride, stride)
    assert C==in_planes
    oc = out_planes
    oh = (H+padding[0]*2-Kh*dilation[0]+dilation[0]-1)//stride[0]+1
    ow = (W+padding[1]*2-Kw*dilation[1]+dilation[1]-1)//stride[1]+1

    if w_ is None:
        assert 0
    else:
        w = w_

    xx = x.reindex([N,G,oc//G,CpG,oh,ow,Kh,Kw], [
        'i0', # Nid
        f'i1*{CpG}+i3', # Gid
        f'i4*{stride[0]}-{padding[0]}+i6*{dilation[0]}', # Hid+Khid
        f'i5*{stride[1]}-{padding[1]}+i7*{dilation[1]}', # Wid+KWid
    ])
    # w: [oc, CpG, Kh, Kw]
    ww = w.reindex([N, G, oc//G, CpG, oh, ow, Kh, Kw], [
        f'i1*{oc//G}+i2',
        'i3',
        'i6',
        'i7'
    ])

    yy = xx*ww
    y = yy.reindex_reduce('add', [N, oc, oh, ow], [
        'i0',
        f'i1*{oc//G}+i2',
        'i4',
        'i5'
    ])
    return y


def test_nchw(x, w, stride, padding, dilation, groups):
    _, in_planes, _, _ = x.shape
    out_planes, _, kernel_size, _ = w.shape
    return conv_nchw(x, in_planes, out_planes, kernel_size, padding, stride=stride, dilation=dilation, groups=groups, w_=w)


def check_forward(xshape, wshape, stride, padding, dilation, groups, use_cuda, nhwc):
    assert nhwc == 0
    test_func = test_nchw

    # only check cudnn
    with jt.log_capture_scope(use_cuda=use_cuda, enable_tuner=1,
        log_v=10, log_vprefix="op.cc=100,conv_tuner=1000"
    ) as raw_log:
        x = jt.random(xshape)
        w = jt.random(wshape)
        y = test_func(x, w, stride, padding, dilation, groups)
        y.sync()
    with jt.flag_scope(use_cuda=0, enable_tuner=0):
        cy = test_func(x, w, stride, padding, dilation, groups)
        cy.sync()

    logs = find_log_with_re(raw_log, "(Jit op key (not )?found: .*conv.*)")
    assert len(logs)==1
    assert np.allclose(y.data, cy.data)


def check_backward(xshape, wshape, stride, padding, dilation, groups, use_cuda, nhwc):
    assert nhwc == 0
    test_func = test_nchw

    # only check cudnn
    with jt.log_capture_scope(use_cuda=use_cuda, enable_tuner=1,
        log_v=10, log_vprefix="op.cc=100,conv_tuner=1000"
    ) as raw_log:
        x = jt.random(xshape)
        w = jt.random(wshape)
        y = test_func(x, w, stride, padding, dilation, groups)
        y.sync()
        dx, dw = jt.grad(y, [x, w])
        jt.sync([y, dx, dw])
    with jt.flag_scope(use_cuda=0, enable_tuner=0, compile_options={"test":233}):
        cy = test_func(x, w, stride, padding, dilation, groups)
        cdx, cdw = jt.grad(cy, [x, w])
        jt.sync([cy, cdx, cdw])

    logs = find_log_with_re(raw_log, "(Jit op key (not )?found: .*conv.*)")
    assert len(logs)==3
    assert np.allclose(y.data, cy.data)
    assert np.allclose(dw.data, cdw.data, 1e-3), (dw.data, cdw.data, np.abs(dw.data - cdw.data).max())
    assert np.allclose(dx.data, cdx.data, 1e-3), (dx.data, cdx.data, np.abs(dx.data - cdx.data).max())


class TestGroupConvTuner(unittest.TestCase):
    @unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
    def test_forward_cuda(self):
        for groups in [2, 4, 8]:
            check_forward([10,8,100,100], [8,8//groups,3,3], 1, 0, 1, groups, 1, False)
            check_forward([10,8,40,50], [16,8//groups,5,5], 1, 1, 2, groups, 1, False)
            check_forward([10,8,40,50], [16,8//groups,4,4], 3, 1, 3, groups, 1, False)

    def test_forward(self):
        for groups in [2, 4, 8]:
            check_forward([10,8,100,100], [8,8//groups,3,3], 1, 0, 1, groups, 0, False)
            check_forward([10,8,40,50], [16,8//groups,5,5], 1, 1, 2, groups, 0, False)
            check_forward([10,8,40,50], [16,8//groups,4,4], 3, 1, 3, groups, 0, False)

    @unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
    def test_backward_cuda(self):
        for groups in [2, 4, 8]:
            check_backward([10,8,100,100], [8,8//groups,3,3], 1, 0, 1, groups, 1, False)
            check_backward([10,8,40,50], [16,8//groups,5,5], 1, 1, 2, groups, 1, False)
            check_backward([10,8,40,50], [16,8//groups,4,4], 3, 1, 3, groups, 1, False)

    def test_backward(self):
        for groups in [2, 4, 8]:
            check_backward([10,8,100,100], [8,8//groups,3,3], 1, 0, 1, groups, 0, False)
            check_backward([10,8,40,50], [16,8//groups,5,5], 1, 1, 2, groups, 0, False)
            check_backward([10,8,40,50], [16,8//groups,4,4], 3, 1, 3, groups, 0, False)
            
        
if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from .test_core import expect_error
from .test_grad import ngrad
from .test_cuda import test_cuda

def all_eq(x, y):
    if len(x.shape) == 0: x = np.array([x])
    if len(y.shape) == 0: y = np.array([y])
    convert = lambda x: x.astype("uint8") if x.dtype=="bool" else x
    x = convert(x)
    y = convert(y)
    if str(x.dtype).startswith("float"):
        return str(y.dtype).startswith("float") and x.shape == y.shape and (x==y).all()
    return x.dtype == y.dtype and x.shape == y.shape and np.testing.assert_allclose(x, y)

def check(op, *args):
    x = eval(f"np.{op}(*args)")
    y = eval(f"jt.{op}(*args).data")
    all_eq(x, y)

class TestBinaryOp(unittest.TestCase):
    def test_binary_op(self):
        assert np.all(jt.binary(1,2,'maximum').data == 2)
        assert np.all(jt.binary([[1,2]],[[3,4]],'add').data == [[4,6]])
        assert np.all(jt.less(1,2).data)
        assert jt.less(1,2).data.dtype == "bool"
        x = (jt.array(1) << jt.array(3)).data
        assert (x == 8).all()
        x = (jt.array(2) ** jt.array(3)).data
        assert (x == 8).all()
        a = np.array([1,2,3])
        b = np.array([7,10,13])
        check("logical_and", a, b)
        check("logical_or", a, b)
        check("logical_xor", a, b)
        check("bitwise_and", a, b)
        check("bitwise_or", a, b)
        check("bitwise_xor", a, b)
        
    def test_i(self):
        def check(op, a, b):
            if isinstance(a, list):
                a = np.array(a)
                b = np.array(b)
            if jt.flags.use_cuda and op == "@":
                return
            if op=="@":
                a = np.float32(a)
                b = np.float32(b)
            ja = jt.array(a)
            jb = jt.array(b)
            exec(f"ja {op}= jb")
            ja = ja.fetch_sync()
            
            if op == "@":
                # numpy do not support @=
                a = np.array(a) @ np.array(b)
            else:
                a = eval(f"a {op} b")
                a = np.float32(a)
            ja = np.float32(ja)
            
            all_eq(ja, a)
        check("+", 5, 2)
        check("-", 5, 2)
        check("*", 5, 2)
        check("/", 5, 2)
        check("//", 5, 2)
        # check("@", [[5]], [[2]])
        check("%", 5, 2)
        check("**", 5, 2)
        check("<<", 5, 2)
        check(">>", 5, 2)
        check("&", 5, 2)
        check("^", 5, 2)
        check("|", 5, 2)
        
        check("+", [5.0,6.0], [2.0,3.0])
        check("-", [5.0,6.0], [2.0,3.0])
        check("*", [5.0,6.0], [2.0,3.0])
        check("/", [5.0,6.0], [2.0,3.0])
        check("//", [5.0,6.0], [2.0,3.0])
        check("@", [[5,6],[7,8]], [[2,3],[4,5]])
        check("%", [5.0,6.0], [2.0,3.0])
        check("**", [5.0,6.0], [2.0,3.0])
        
    def test_r(self):
        def check(op, a, b):
            a = np.array(a)
            b = np.array(b)
            if jt.flags.use_cuda and op == "@":
                return
            jb = jt.array(b)
            jc = eval(f"a {op} jb").data

            
            if op == "@":
                # numpy do not support @=
                a = np.array(a) @ np.array(b)
            else:
                a = eval(f"a {op} b")
                a = np.array(a)
            
            all_eq(jc, a)
        check("+", 5, 2)
        check("-", 5, 2)
        check("*", 5, 2)
        check("/", 5, 2)
        check("//", 5, 2)
        # check("@", [[5]], [[2]])
        check("%", 5, 2)
        check("**", 5, 2)
        check("<<", 5, 2)
        check(">>", 5, 2)
        check("&", 5, 2)
        check("^", 5, 2)
        check("|", 5, 2)
        
    def test_grad(self):
        ops = ["+", "-", "*", "/", "**"]
        np.random.seed(3)
        a = np.random.rand(10)
        b = np.random.rand(10)
        c = np.random.rand(10)
        tol = 1e-2 if jt.flags.amp_reg & 2 else 1e-4
        for op in ops:
            func = lambda x: eval(f"((x[0]{op}x[1])*x[2]).sum()")
            x, grads = ngrad(func, [a,b,c], 1e-8)
            ja = jt.array(a).name("ja")
            jb = jt.array(b).name("jb")
            jc = jt.array(c).name("jc")
            jx = eval(f"(ja{op}jb)*jc")
            jgrads = jt.grad(jx, [ja,jb,jc])
            for jd, nd in zip(jgrads, grads):
                np.testing.assert_allclose(jd.data, nd, atol=tol, rtol=tol)

    def test_mod_float(self):
        a = jt.random((10,))
        b = jt.random((10,))
        c = a % b
        assert np.allclose(c.data, a.data % b.data)
        a = jt.random((10,), 'float64')
        b = jt.random((10,), 'float64')
        c = a % b
        assert np.allclose(c.data, a.data % b.data, a.data, b.data)
        if jt.flags.amp_reg & 2: return
        a = jt.random((10,)) * 1000
        b = (jt.random((10,)) * 10).int() + 1
        c = a % b
        assert np.allclose(c.data, a.data % b.data), (c.data, a.data%b.data)

    def test_mod_grad(self):
        a = jt.random((100,))
        b = jt.random((100,))
        c = a % b
        da, db = jt.grad(c, [a, b])
        np.testing.assert_allclose(da.data, 1)
        np.testing.assert_allclose(db.data, -np.floor(a.data/b.data))

    def test_mod_negtive(self):
        a = jt.random((100,)) - 0.5
        b = jt.random((100,)) - 0.5
        c = a % b
        nc = a.data % b.data
        np.testing.assert_allclose(c.data, nc.data, atol=1e-5, rtol=1e-5)
    
    def test_pow(self):
        # win cuda 10.2 cannot pass
        a = jt.random((100,))
        b = a**3
        b.sync()

    def test_binary_op_bool(self):
        a = np.array([0,1,0,1]).astype(bool)
        b = np.array([0,1,1,0]).astype(bool)
        c = np.array([1,1,1,1]).astype(bool)
        check("add", a, b)
        all_eq(np.logical_xor(a, b), jt.subtract(a, b).data)
        check("multiply", a, b)
        check("logical_and", a, b)
        check("logical_or", a, b)
        check("logical_xor", a, b)
        check("bitwise_and", a, b)
        check("bitwise_or", a, b)
        check("bitwise_xor", a, b)
        check("divide", a, c)
        check("floor_divide", a, c)
        check("mod", a, c)


class TestBinaryOpCuda(TestBinaryOp, test_cuda(2)):
    pass

class TestBinaryOpCpuFp16(TestBinaryOp):
    def setUp(self):
        jt.flags.amp_reg = 2 | 4 | 8 | 16
    def tearDown(self):
        jt.flags.amp_reg = 0

@unittest.skipIf(not jt.has_cuda, "no cuda found")
class TestBinaryOpCudaFp16(TestBinaryOp):
    def setUp(self):
        jt.flags.amp_reg = 2 | 4 | 8 | 16
        jt.flags.use_cuda = 1
    def tearDown(self):
        jt.flags.amp_reg = 0
        jt.flags.use_cuda = 0

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers:
#   Dun Liang <randonlang@gmail.com>.
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
from socketserver import ThreadingTCPServer
import socket
import os
import sys
import threading

key_queue = {}

def handle_connect(req:socket.socket, c_addr, server):
    print("get connect", c_addr, req)
    skey = req.recv(1024).decode()
    print("get skey", skey)
    with lock:
        if skey not in key_queue:
            key_queue[skey] = []
        queue = key_queue[skey]
        queue.append(req)

        req.send(str(len(queue)-1).encode())
    while True:
        buf = req.recv(1024).decode()
        print(buf)
        with lock:
            if len(buf) == 0:
                for i,r in enumerate(queue):
                    if r is req:
                        for j in range(i+1, len(queue)):
                            queue[j].send(str(j-1).encode())
                        del queue[i]
                        print("queue size", len(queue))
                        break
                break
        

def wait_queue():
    global s
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect(("127.0.0.1", 8900))
    s.sendall(skey.encode())
    while True:
        buf = s.recv(1024).decode()
        if len(buf) == 0:
            print("Cannot connect to queue server, please report this issue to admin.")
            sys.exit(1)
        if buf == '0':
            print("Begin")
            os.system(f"sleep {os.environ.get('SWAIT', '60')} && bash -c ' if kill -9 {os.getpid()} 2>/dev/null; then echo Timeout; fi; ' &")
            break
        else:
            print("Pending", buf)



if "SKEY" in os.environ:
    skey = os.environ["SKEY"]
    wait_queue()
else:
    lock = threading.Lock()
    server = ThreadingTCPServer(("127.0.0.1", 8900), handle_connect)
    server.serve_forever()



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Guoye Yang <498731903@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from jittor import compile_extern
if jt.has_cuda:
    from jittor.compile_extern import cublas_ops, cudnn_ops, cub_ops
else:
    cublas_ops = cudnn_ops = cub_ops = None


def test_forward(shape, dim=None):
    x = jt.random(shape)
    y = jt.numpy_cumsum(x, dim=dim)
    y_ = jt.cub_cumsum(x, dim=dim)
    assert(np.allclose(y.data, y_.data))

def test_backward(shape, dim=None):
    x = jt.random(shape)
    z = jt.random(shape)

    y = jt.numpy_cumsum(x, dim=dim)
    loss = (y * z).sum()
    grad = jt.grad(loss, x)
    
    y_ = jt.cub_cumsum(x, dim=dim)
    loss_ = (y_ * z).sum()
    grad_ = jt.grad(loss_, x)
    assert(np.allclose(grad.data, grad_.data))

class TestCubCumsumOp(unittest.TestCase):
    def setUp(self):
        self.is_reversed = False

    @unittest.skipIf(cub_ops==None, "Not use cub, Skip")
    @jt.flag_scope(use_cuda=1)
    def test_1d(self):
        test_forward([20])
        test_forward([6007])
        test_forward([6007], 0)
        test_forward([6007], -1)

    @unittest.skipIf(cub_ops==None, "Not use cub, Skip")
    @jt.flag_scope(use_cuda=1)
    def test_1d_backward(self):
        test_backward([20])
        test_backward([6007])
        test_backward([6007], 0)
        test_backward([6007], -1)

    @unittest.skipIf(cub_ops==None, "Not use cub, Skip")
    @jt.flag_scope(use_cuda=1)
    def test_2d(self):
        test_forward([5,5])
        test_forward([2000, 6007])
        test_forward([2000, 6007], 1)
        test_forward([2000, 6007], -1)

    @unittest.skipIf(cub_ops==None, "Not use cub, Skip")
    @jt.flag_scope(use_cuda=1)
    def test_2d_backward(self):
        test_backward([5,5])
        test_backward([2000, 6007])
        test_backward([2000, 6007], 1)
        test_backward([2000, 6007], -1)

    @unittest.skipIf(cub_ops==None, "Not use cub, Skip")
    @jt.flag_scope(use_cuda=1)
    def test_nd(self):
        test_forward([5,6,7,8], 0)
        test_forward([5,6,7,8], 1)
        test_forward([5,6,7,8], 2)
        test_forward([5,6,7,8], 3)
        test_forward([5,6,7,8], -1)
        test_forward([16,14,14,4096], 0)
        test_forward([16,14,14,4096], 1)
        test_forward([16,14,14,4096], 2)
        test_forward([16,14,14,4096], 3)
        test_forward([16,14,14,4096], -1)
        test_forward([16,14,14,4095], 3)

    @unittest.skipIf(cub_ops==None, "Not use cub, Skip")
    @jt.flag_scope(use_cuda=1)
    def test_nd_backward(self):
        test_backward([5,6,7,8], 0)
        test_backward([5,6,7,8], 1)
        test_backward([5,6,7,8], 2)
        test_backward([5,6,7,8], 3)
        test_backward([5,6,7,8], -1)
        test_backward([16,14,14,4096], 0)
        test_backward([16,14,14,4096], 1)
        test_backward([16,14,14,4096], 2)
        test_backward([16,14,14,4096], 3)
        test_backward([16,14,14,4096], -1)
        test_backward([16,14,14,4095], 3)

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import sys
import os
import jittor as jt
import unittest
import time
import numpy as np
from .test_reorder_tuner import simple_parser
from .test_log import find_log_with_re

class TestMatmulTuner(unittest.TestCase):
    def test_matmul_tuner(self):
        n,m,k = 10,10,10
        a = jt.random([n,m])
        b = jt.random([m,k])
        with jt.log_capture_scope(
            log_v=0, log_vprefix="tuner_manager=100,var_relay=100",
            compile_options={"test_matmul_tuner":1}
        ) as rawlogs:
            c = a.broadcast([n,m,k], [2]) * b.broadcast([n,m,k], [0])
            c = c.sum(1)
            jc = c.numpy()
            nc = np.matmul(a.numpy(), b.numpy())
            assert (np.abs(jc-nc)<1e-3).all()
        logs = find_log_with_re(rawlogs, 
            "Run tuner matmul: confidence\\((.*)\\) candidates\\((.*)\\)$")
        assert len(logs) == 1
        assert logs[0][0] == "20", "confidence of reorder should be 20"
        candidates = simple_parser(logs[0][1])
        assert candidates == {"relay0":[1,0]}, candidates
        logs = find_log_with_re(rawlogs, r"get_relay_src([\s\S]*)")
        assert len(logs)==1
        assert "@relay_op" in logs[0]


if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest, os
import jittor as jt
from jittor import LOG
import sys
import jittor_utils as jit_utils

dirname = os.path.join(jt.flags.jittor_path, "notebook")
notebook_dir = os.path.join(jit_utils.home(), ".cache","jittor","notebook")
tests = []
for mdname in os.listdir(dirname):
    if not mdname.endswith(".src.md"): continue
    # temporary disable model_test
    if "GAN" in mdname: continue
    tests.append(mdname[:-3])

try:
    jt.compiler.run_cmd("ipython --help")
    has_ipython = True
except:
    has_ipython = False

def test(name):
    LOG.i(f"Run test {name} from {dirname}")
    ipynb_name = os.path.join(notebook_dir, name+".ipynb")
    jt.compiler.run_cmd("ipython "+ipynb_name)

def init():
    cmd = sys.executable+" "+os.path.join(dirname, "md_to_ipynb.py")
    LOG.i("init notebooks:", cmd)
    jt.compiler.run_cmd(cmd)

src = """class TestNodebooks(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        init()
"""
for name in tests:
    src += f"""
    @unittest.skipIf(not has_ipython, "No IPython found")
    def test_{name.replace(".src","")}(self):
        test("{name}")
    """

LOG.vvv("eval src\n"+src)
exec(src)

if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved.
# Maintainers:
#     Haoyang Peng <2247838039@qq.com>
#     Guowei Yang <471184555@qq.com>
#     Dun Liang <randonlang@gmail.com>.
#
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
import numpy as np
import unittest

try:
    import torch
    from torch.autograd import Variable
    import autograd.numpy as anp
    from autograd import jacobian

    has_autograd = True
except:
    has_autograd = False

cupy = None
try:
    import cupy
except:
    pass

@unittest.skipIf(not has_autograd, "No autograd found.")
class TestEinsum(unittest.TestCase):
    def test_einsum_ijjk(self):
        for i in range(30):
            string = "ij,jk->ik"
            tn, tm = np.random.randn(3, 3).astype('float32'), np.random.randn(3, 3).astype('float32')
            x = jt.array(tn)
            y = jt.array(tm)
            t_x = torch.from_numpy(tn)
            t_y = torch.from_numpy(tm)
            t_x = Variable(t_x, requires_grad=True)
            t_y = Variable(t_y, requires_grad=True)
            jq = jt.linalg.einsum(string, x, y)
            tq = torch.einsum(string, t_x, t_y)
            np.testing.assert_allclose(jq.data, tq.detach().numpy(), rtol=1e-4, atol=1e-6)
            gq = jt.grad(jq, x).data
            gr = jt.grad(jq, y).data
            tgq = torch.autograd.grad(tq, t_x, torch.ones_like(tq), retain_graph=True)
            tgr = torch.autograd.grad(tq, t_y, torch.ones_like(tq))
            np.testing.assert_allclose(gq, tgq[0].numpy(), rtol=1e-4, atol=1e-6)
            np.testing.assert_allclose(gr, tgr[0].numpy(), rtol=1e-4, atol=1e-6)
    
    def test_einsum_ii(self):
        for i in range(30):
            string = "ij->i"
            tn, tm = np.random.randn(3, 3).astype('float32'), np.random.randn(3, 3).astype('float32')
            x = jt.array(tn)
            # x = x.reindex([2, 2, x.shape[0], x.shape[1]], ["i2", "i3"])
            t_x = torch.from_numpy(tn)
            t_x = Variable(t_x, requires_grad=True)
            jq = jt.linalg.einsum(string, x)
            tq = torch.einsum(string, t_x)
            np.testing.assert_allclose(jq.data, tq.detach().numpy(), rtol=1e-4, atol=1e-6)
            gq = jt.grad(jq, x).data
            tgq = torch.autograd.grad(tq, t_x, torch.ones_like(tq))
            np.testing.assert_allclose(gq, tgq[0].numpy(), rtol=1e-4, atol=1e-6)
    
    def test_einsum_multi(self):
       for i in range(30):
            string = "ij,ijk,jk->ik"
            tn, tm, tk = np.random.randn(3, 4).astype('float32'), np.random.randn(3, 4, 5).astype('float32'), np.random.randn(4, 5).astype('float32')
            x = jt.array(tn)
            y = jt.array(tm)
            z = jt.array(tk)
            # x = x.reindex([2, 2, x.shape[0], x.shape[1]], ["i2", "i3"])
            t_x = torch.from_numpy(tn)
            t_y = torch.from_numpy(tm)
            t_z = torch.from_numpy(tk)
            t_x = Variable(t_x, requires_grad=True)
            t_y = Variable(t_y, requires_grad=True)
            t_z = Variable(t_z, requires_grad=True)
            jq = jt.linalg.einsum(string, x, y, z)
            tq = torch.einsum(string, t_x, t_y, t_z)
            np.testing.assert_allclose(jq.data, tq.detach().numpy(), rtol=1e-4, atol=1e-6)
            gq = jt.grad(jq, x).data
            gr = jt.grad(jq, y).data
            gz = jt.grad(jq, z).data
            tgq = torch.autograd.grad(tq, t_x, torch.ones_like(tq), retain_graph=True)
            tgr = torch.autograd.grad(tq, t_y, torch.ones_like(tq), retain_graph=True)
            tgz = torch.autograd.grad(tq, t_z, torch.ones_like(tq), retain_graph=True)
            np.testing.assert_allclose(gq, tgq[0].numpy(), rtol=1e-4, atol=1e-6)
            np.testing.assert_allclose(gr, tgr[0].numpy(), rtol=1e-4, atol=1e-6)
            np.testing.assert_allclose(gz, tgz[0].numpy(), rtol=1e-4, atol=1e-6)


@unittest.skipIf(not jt.compiler.has_cuda or cupy is None, "No CUDA found")
class TestCudaEinsum(TestEinsum):
    def setUp(self):
        jt.flags.use_cuda = 1
    def tearDown(self):
        jt.flags.use_cuda = 0

if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
import os

from jittor.test.misc import superglue
from jittor.test.misc.superglue import SuperGlue
import time

@jt.flag_scope(use_cuda=1)
def main():
    global superglue
    superglue.split_size = int(os.environ.get("split_size", "12"))
    # superglue.split_size = 1000000

    batch = 30
    num = 2000
    dim = 128

    # jt.display_memory_info()
    # os.system("nvidia-smi")
    # breakpoint()

    with jt.no_grad():

        config = {
            'superglue': {
                'sinkhorn_iterations': 25,
                'match_threshold': 0.01,
                'keypoint_position_dim': 2,
                'descriptor_dim': dim,
                'use_dual_softmax': True,
                'GNN_layers': ['self', 'cross'] * 9,
            }
        }

        superglue = SuperGlue(config.get('superglue', {}))

        superglue.eval()

        data = {
            'keypoints0': jt.rand((batch, num, 2), dtype=jt.float),
            'keypoints1': jt.rand((batch, num, 2), dtype=jt.float),
            'shape0': jt.rand((batch, 2), dtype=jt.float),
            'shape1': jt.rand((batch, 2), dtype=jt.float),
            'descriptors0': jt.rand((batch, dim, num), dtype=jt.float),
            'descriptors1': jt.rand((batch, dim, num), dtype=jt.float),
            'scores0': jt.rand((batch, num), dtype=jt.float),
            'scores1': jt.rand((batch, num), dtype=jt.float),
            'all_matches': jt.randint(0, num, (batch, num, 2), dtype=jt.int),
            'return_match': False,
            # 'match_num': match_num
        }

        use_fp16 = int(os.environ.get("use_fp16", "0"))
        if use_fp16:
            jt.flags.amp_reg = 2
            for k,v in data.items():
                if isinstance(v, jt.Var) and v.dtype == "float32":
                    v.assign(v.float16())
            for v in superglue.parameters():
                if v.dtype == "float32":
                    v.assign(v.float16())
            jt.sync_all(True)

        import pickle
        jt.sync_all(True)
        for x in range(5):
            print(x)
            jt.gc()
            x = superglue(data)['loss']
            x.sync()
            jt.display_memory_info()
            # os.system("nvidia-smi")
            # breakpoint()
            # print(data)
            # print(x)
        
        # with open("/tmp/record.pkl", "wb") as f:
        #     pickle.dump([data, x], f, pickle.HIGHEST_PROTOCOL)

        # with jt.flag_scope(trace_py_var=3, profile_memory_enable=1):
        #     x = superglue(data)['loss']
        #     x.sync()
        #     jt.get_max_memory_treemap()
        # exit(0)

        jt.sync_all(True)
        time0 = time.time()
        jt.flags.profiler_enable = int(os.environ.get("profiler", "0"))

        for x in range(20):
            print(x)
            # jt.display_memory_info()
            x = superglue(data)['loss']
            x.sync()
            # print(x)

        jt.sync_all(True)
        time1 = time.time()
        print("avg time:", (time1 - time0) / 20)
        return (time1 - time0) / 20


class TestSuperglue(unittest.TestCase):
    def test(self):
        if not jt.has_cuda: return
        t1 = main()
        os.environ["use_fp16"] = "1"
        t2 = main()
        os.environ["use_fp16"] = "0"
        assert t1*0.55 > t2

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Zheng-Ning Liu <lzhengning@gmail.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************


import unittest
import jittor as jt
import numpy as np


class TestRepeatOp(unittest.TestCase):
    def test_repeat(self):
        np_a = np.arange(5)
        jt_a = jt.array(np_a)

        np_b = np.tile(np_a, (2, 3))
        jt_b = jt.repeat(jt_a, (2, 3))

        assert np.allclose(np_b, jt_b.data)

        np_b = np.tile(np_a, (2, 3, 1))
        jt_b = jt.repeat(jt_a, (2, 3, 1))

        assert np.allclose(np_b, jt_b.data)

        np_a = np.arange(24).reshape(2, 3, 4)
        jt_a = jt.array(np_a)

        np_b = np.tile(np_a, (2, 3))
        jt_b = jt.repeat(jt_a, (2, 3))

        assert np.allclose(np_b, jt_b.data)


    def test_highdim(self):
        np_a = np.arange(64).reshape(2, 2, 2, 2, 2, 2)
        jt_a = jt.array(np_a)

        np_b = np.tile(np_a, (2, 3))
        jt_b = jt.repeat(jt_a, (2, 3))

        assert np.allclose(np_b, jt_b.data)

        np_b = np.tile(np_a, (2, 1, 1, 3))
        jt_b = jt.repeat(jt_a, (2, 1, 1, 3))

        assert np.allclose(np_b, jt_b.data)

        np_b = np.tile(np_a, (2, 1, 1, 1, 3, 1))
        jt_b = jt.repeat(jt_a, (2, 1, 1, 1, 3, 1))

        assert np.allclose(np_b, jt_b.data)

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved.
# Maintainers:
#     Haoyang Peng <2247838039@qq.com>
#     Guowei Yang <471184555@qq.com>
#     Dun Liang <randonlang@gmail.com>.
#
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
import numpy as np
import unittest

try:
    import torch
    from torch.autograd import Variable
    import autograd.numpy as anp
    from autograd import jacobian

    has_autograd = True
except:
    has_autograd = False

@unittest.skipIf(not has_autograd, "No autograd found.")
class TestHistc(unittest.TestCase):
    def test_histc(self):
        for i in range(30):
            inputs = np.random.uniform(0,10,(40,40))
            tn, tm = np.random.randn(3, 3).astype('float32'), np.random.randn(3, 3).astype('float32')
            x = jt.array(inputs)
            t_x = torch.from_numpy(inputs)
            if i % 2:
                min = max = 0
            else:
                min = (inputs.min() + inputs.max()) / 3
                max = (inputs.min() + inputs.max()) / 3 * 2
            joup = jt.histc(x, bins=i+1, min=min, max=max)
            toup = torch.histc(t_x, bins=i+1, min=min, max=max)
            np.testing.assert_allclose(joup.data, toup.cpu().numpy(), rtol=1e-4, atol=1e-6)

if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved.
# Maintainers:
#     Haoyang Peng <2247838039@qq.com>
#     Dun Liang <randonlang@gmail.com>.
#
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
import numpy as np
import unittest

try:
    import torch
    from torch.autograd import Variable
    has_autograd = True
except:
    has_autograd = False

@unittest.skipIf(not has_autograd, "No autograd found.")
class TestDigamma(unittest.TestCase):
    def test_digamma(self):
        for i in range(30):
            nx = np.random.uniform(0, 1, (32, 32))
            x = jt.array(nx)
            tx = torch.autograd.Variable(torch.tensor(nx, dtype=torch.float32), requires_grad=True)
            dx = jt.digamma.apply(x)
            tdx = torch.digamma(tx)
            np.testing.assert_allclose(dx.data, tdx.detach().numpy(), rtol=1e-4, atol=1e-6)
            jgdx = jt.grad(dx, x)
            tgdx = torch.autograd.grad(tdx, tx, torch.ones_like(tx))[0]
            np.testing.assert_allclose(jgdx.data, tgdx.detach().numpy(), rtol=1e-4, atol=1e-6)

@unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
class TestCudaDigamma(TestDigamma):
    def setUp(self):
        jt.flags.use_cuda = 1
    def tearDown(self):
        jt.flags.use_cuda = 0

if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved.
# Maintainers: DongYang Li <lidongyang2001@gmail.com>.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
from collections import namedtuple
import tempfile
import pickle
import itertools
from jittor.einops.einops import (rearrange, reduce, _enumerate_directions, _reductions)
from jittor.einops import EinopsError
import jittor as jt
import numpy
import unittest

# tests/__init__.py
import os
from jittor.einops import _backends
import warnings

flag_to_bool = {
    '': False,
    '0': False,
    '1': True,
}


def collect_test_backends(symbolic=False, layers=False):
    """
    :param symbolic: symbolic or imperative frameworks?
    :param layers: layers or operations?
    :return: list of backends satisfying set conditions
    """
    if not symbolic:
        if not layers:
            backend_types = [
                _backends.NumpyBackend,
                _backends.JittorBackend,
            ]
        else:
            backend_types = [
                _backends.JittorBackend,
            ]
    else:
        backend_types = []
    result = []
    for backend_type in backend_types:
        try:
            result.append(backend_type())
        except ImportError:
            # problem with backend installation fails a specific test function,
            # but will be skipped in all other test cases
            warnings.warn('backend could not be initialized for tests: {}'.format(backend_type))
    return result


# test/test_ops.py

imp_op_backends = collect_test_backends(symbolic=False, layers=False)

# test/test_layer.py


class TestSlice(unittest.TestCase):

    def test_anonymous_axes(self):
        x = numpy.arange(1 * 2 * 4 * 6).reshape([1, 2, 4, 6])
        for pattern, axis_dimensions in test_cases_repeat_anonymous:
            check_reversion(x, pattern, **axis_dimensions)

    def test_repeat_imperatives(self):
        x = numpy.arange(2 * 3 * 5).reshape([2, 3, 5])
        for backend in imp_op_backends:
            print('Repeat tests for ', backend.framework_name)

            for pattern, axis_dimensions in repeat_test_cases:
                expected = reduce(x, pattern, reduction='repeat', **axis_dimensions)
                converted = backend.from_numpy(x)
                repeated = reduce(converted, pattern, reduction='repeat', **axis_dimensions)
                result = backend.to_numpy(repeated)
                assert numpy.array_equal(result, expected)

    def test_repeat_numpy(self):
        # check repeat vs reduce. Repeat works ok if reverse reduction with min and max work well
        x = numpy.arange(2 * 3 * 5).reshape([2, 3, 5])
        x1 = reduce(x, 'a b c -> copy a b c ', reduction='repeat', copy=1)
        assert numpy.array_equal(x[None], x1)
        for pattern, axis_dimensions in repeat_test_cases:
            check_reversion(x, pattern, **axis_dimensions)

    def test_tiling_imperatives(self):
        for backend in imp_op_backends:
            print('Tiling tests for ', backend.framework_name)
            input = numpy.arange(2 * 3 * 5, dtype='int64').reshape([2, 1, 3, 1, 5])
            test_cases = [
                (1, 1, 1, 1, 1),
                (1, 2, 1, 3, 1),
                (3, 1, 1, 4, 1),
            ]
            for repeats in test_cases:
                expected = numpy.tile(input, repeats)
                converted = backend.from_numpy(input)
                repeated = backend.tile(converted, repeats)
                result = backend.to_numpy(repeated)
                assert numpy.array_equal(result, expected)

    def test_gradients_imperatives(self):
        # lazy - just checking reductions
        for reduction in _reductions:
            x = numpy.arange(1, 1 + 2 * 3 * 4).reshape([2, 3, 4]).astype('float32')
            results = {}
            for backend in imp_op_backends:
                y0 = backend.from_numpy(x)
                if not 'jittor' in backend.framework_name and not hasattr(y0, 'grad'):
                    continue
                y1 = reduce(y0, 'a b c -> c a', reduction=reduction)
                y2 = reduce(y1, 'c a -> a c', reduction=reduction)
                y3 = reduce(y2, 'a (c1 c2) -> a', reduction=reduction, c1=2)
                y4 = reduce(y3, '... -> ', reduction=reduction)
                if 'jittor' in backend.framework_name:
                    grad = backend.jittor.grad(y4, y0)
                else:
                    y4.backward()
                    grad = y0.grad
                results[backend.framework_name] = backend.to_numpy(grad)

            print('comparing gradients for', results.keys())
            for name1, grad1 in results.items():
                for name2, grad2 in results.items():
                    assert numpy.allclose(grad1, grad2), [name1, name2, 'provided different gradients']

    def test_concatenations_and_stacking(self):
        for backend in imp_op_backends:
            print('testing shapes for ', backend.framework_name)
            for n_arrays in [1, 2, 5]:
                shapes = [[], [1], [1, 1], [2, 3, 5, 7], [1] * 6]
                for shape in shapes:
                    if (backend.framework_name == 'jittor')\
                            and len(shape) == 0:
                        # jittor stores scalar in 1d array
                        continue
                    arrays1 = [numpy.arange(i, i + numpy.prod(shape)).reshape(shape) for i in range(n_arrays)]
                    arrays2 = [backend.from_numpy(array) for array in arrays1]
                    result0 = numpy.asarray(arrays1)
                    result1 = rearrange(arrays1, '...->...')
                    result2 = rearrange(arrays2, '...->...')
                    assert numpy.array_equal(result0, result1)
                    assert numpy.array_equal(result1, backend.to_numpy(result2))

                    result1 = rearrange(arrays1, 'b ... -> ... b')
                    result2 = rearrange(arrays2, 'b ... -> ... b')
                    assert numpy.array_equal(result1, backend.to_numpy(result2))

    def test_enumerating_directions(self):
        for backend in imp_op_backends:
            print('testing directions for', backend.framework_name)
            for shape in [[], [1], [1, 1, 1], [2, 3, 5, 7]]:
                if (backend.framework_name == 'jittor')\
                        and len(shape) == 0:
                    # jittor stores scalar in 1d array
                    continue
                x = numpy.arange(numpy.prod(shape)).reshape(shape)
                axes1 = _enumerate_directions(x)
                axes2 = _enumerate_directions(backend.from_numpy(x))
                assert len(axes1) == len(axes2) == len(shape)
                for ax1, ax2 in zip(axes1, axes2):
                    ax2 = backend.to_numpy(ax2)
                    assert ax1.shape == ax2.shape
                    assert numpy.allclose(ax1, ax2)

    def test_reduction_with_callable_imperatives(self):
        x_numpy = numpy.arange(2 * 3 * 4 * 5 * 6).reshape([2, 3, 4, 5, 6]).astype('float32')
        x_numpy /= x_numpy.max()

        def logsumexp_jittor(x, tuple_of_axes):
            import jittor as jt
            return jt.nn.logsumexp(x, tuple_of_axes)

        def logsumexp_numpy(x, tuple_of_axes):
            # very naive logsumexp to compare to
            minused = x.max(tuple_of_axes)
            y = x - x.max(tuple_of_axes, keepdims=True)
            y = numpy.exp(y)
            y = numpy.sum(y, axis=tuple_of_axes)
            return numpy.log(y) + minused

        from jittor.einops._backends import JittorBackend, NumpyBackend
        backend2callback = {
            JittorBackend.framework_name: logsumexp_jittor,
            NumpyBackend.framework_name: logsumexp_numpy,
        }

        for backend in imp_op_backends:
            if backend.framework_name not in backend2callback:
                continue

            backend_callback = backend2callback[backend.framework_name]

            x_backend = backend.from_numpy(x_numpy)
            for pattern1, pattern2 in equivalent_reduction_patterns:
                print('Test reduction with callable for ', backend.framework_name, pattern1, pattern2)
                output_numpy = reduce(x_numpy, pattern1, reduction=logsumexp_numpy)
                output_backend = reduce(x_backend, pattern1, reduction=backend_callback)
                assert numpy.allclose(
                    output_numpy,
                    backend.to_numpy(output_backend),
                )

    def test_reduction_stress_imperatives(self):
        for backend in imp_op_backends:
            print('Stress-testing reduction for ', backend.framework_name)
            for reduction in _reductions + ('rearrange',):
                dtype = 'int64'
                coincide = numpy.array_equal
                if reduction in ['mean', 'prod']:
                    dtype = 'float64'
                    coincide = numpy.allclose
                for n_axes in range(11):
                    shape = numpy.random.randint(2, 4, size=n_axes)
                    permutation = numpy.random.permutation(n_axes)
                    skipped = 0 if reduction == 'rearrange' else numpy.random.randint(n_axes + 1)
                    left = ' '.join('x' + str(i) for i in range(n_axes))
                    right = ' '.join('x' + str(i) for i in permutation[skipped:])
                    pattern = left + '->' + right
                    x = numpy.arange(1, 1 + numpy.prod(shape), dtype=dtype).reshape(shape)
                    if reduction == 'prod':
                        x /= x.mean()  # to avoid overflows
                    result1 = reduce(x, pattern, reduction=reduction)
                    result2 = x.transpose(permutation)
                    if skipped > 0:
                        result2 = getattr(result2, reduction)(axis=tuple(range(skipped)))
                    assert coincide(result1, result2)
                    check_op_against_numpy(backend, x, pattern, reduction=reduction, axes_lengths={}, is_symbolic=False)

    def test_reduction_imperatives(self):
        for backend in imp_op_backends:
            print('Reduction tests for ', backend.framework_name)
            for reduction in _reductions:
                # slight redundancy for simpler order - numpy version is evaluated multiple times
                input = numpy.arange(2 * 3 * 4 * 5 * 6, dtype='int64').reshape([2, 3, 4, 5, 6])
                if reduction in ['mean', 'prod']:
                    input = input / input.astype('float64').mean()
                test_cases = [
                    ['a b c d e -> ', {},
                     getattr(input, reduction)()],
                    ['a ... -> ', {},
                     getattr(input, reduction)()],
                    ['(a1 a2) ... (e1 e2) -> ', dict(a1=1, e2=2),
                     getattr(input, reduction)()],
                    ['a b c d e -> (e c) a', {},
                     getattr(input, reduction)(axis=(1, 3)).transpose(2, 1, 0).reshape([-1, 2])],
                    ['a ... c d e -> (e c) a', {},
                     getattr(input, reduction)(axis=(1, 3)).transpose(2, 1, 0).reshape([-1, 2])],
                    ['a b c d e ... -> (e c) a', {},
                     getattr(input, reduction)(axis=(1, 3)).transpose(2, 1, 0).reshape([-1, 2])],
                    ['a b c d e -> (e c a)', {},
                     getattr(input, reduction)(axis=(1, 3)).transpose(2, 1, 0).reshape([-1])],
                    ['(a a2) ... -> (a2 a) ...', dict(a2=1),
                     input],
                ]
                for pattern, axes_lengths, expected_result in test_cases:
                    result = reduce(backend.from_numpy(input.copy()), pattern, reduction=reduction, **axes_lengths)
                    result = backend.to_numpy(result)
                    assert numpy.allclose(result, expected_result)

    def test_rearrange_permutations_numpy(self):
        # tests random permutation of axes against two independent numpy ways
        for n_axes in range(1, 10):
            input = numpy.arange(2 ** n_axes).reshape([2] * n_axes)
            permutation = numpy.random.permutation(n_axes)
            left_expression = ' '.join('i' + str(axis) for axis in range(n_axes))
            right_expression = ' '.join('i' + str(axis) for axis in permutation)
            expression = left_expression + ' -> ' + right_expression
            result = rearrange(input, expression)

            for pick in numpy.random.randint(0, 2, [10, n_axes]):
                assert input[tuple(pick)] == result[tuple(pick[permutation])]

        for n_axes in range(1, 10):
            input = numpy.arange(2 ** n_axes).reshape([2] * n_axes)
            permutation = numpy.random.permutation(n_axes)
            left_expression = ' '.join('i' + str(axis) for axis in range(n_axes)[::-1])
            right_expression = ' '.join('i' + str(axis) for axis in permutation[::-1])
            expression = left_expression + ' -> ' + right_expression
            result = rearrange(input, expression)
            assert result.shape == input.shape
            expected_result = numpy.zeros_like(input)
            for original_axis, result_axis in enumerate(permutation):
                expected_result |= ((input >> original_axis) & 1) << result_axis

            assert numpy.array_equal(result, expected_result)

    def test_rearrange_consistency_numpy(self):
        shape = [1, 2, 3, 5, 7, 11]
        x = numpy.arange(numpy.prod(shape)).reshape(shape)
        for pattern in [
            'a b c d e f -> a b c d e f',
            'b a c d e f -> a b d e f c',
            'a b c d e f -> f e d c b a',
            'a b c d e f -> (f e) d (c b a)',
            'a b c d e f -> (f e d c b a)',
        ]:
            result = rearrange(x, pattern)
            assert len(numpy.setdiff1d(x, result)) == 0
            assert result.dtype == x.dtype

        result = rearrange(x, 'a b c d e f -> a (b) (c d e) f')
        assert numpy.array_equal(x.flatten(), result.flatten())

        result = rearrange(x, 'a aa aa1 a1a1 aaaa a11 -> a aa aa1 a1a1 aaaa a11')
        assert numpy.array_equal(x, result)

        result1 = rearrange(x, 'a b c d e f -> f e d c b a')
        result2 = rearrange(x, 'f e d c b a -> a b c d e f')
        assert numpy.array_equal(result1, result2)

        result = rearrange(rearrange(x, 'a b c d e f -> (f d) c (e b) a'), '(f d) c (e b) a -> a b c d e f', b=2, d=5)
        assert numpy.array_equal(x, result)

        sizes = dict(zip('abcdef', shape))
        temp = rearrange(x, 'a b c d e f -> (f d) c (e b) a', **sizes)
        result = rearrange(temp, '(f d) c (e b) a -> a b c d e f', **sizes)
        assert numpy.array_equal(x, result)

        x2 = numpy.arange(2 * 3 * 4).reshape([2, 3, 4])
        result = rearrange(x2, 'a b c -> b c a')
        assert x2[1, 2, 3] == result[2, 3, 1]
        assert x2[0, 1, 2] == result[1, 2, 0]

    def test_ellipsis_ops_imperative(self):
        """ Checking various patterns against numpy """
        x = numpy.arange(2 * 3 * 4 * 5 * 6).reshape([2, 3, 4, 5, 6])
        for is_symbolic in [True, False]:
            for backend in collect_test_backends(symbolic=is_symbolic, layers=False):
                for pattern in identity_patterns + list(itertools.chain(*equivalent_rearrange_patterns)):
                    check_op_against_numpy(backend, x, pattern, axes_lengths={},
                                           reduction='rearrange', is_symbolic=is_symbolic)

                for reduction in ['min', 'max', 'sum']:
                    for pattern in itertools.chain(*equivalent_reduction_patterns):
                        check_op_against_numpy(backend, x, pattern, axes_lengths={},
                                               reduction=reduction, is_symbolic=is_symbolic)

    def test_ellipsis_ops_numpy(self):
        x = numpy.arange(2 * 3 * 4 * 5 * 6).reshape([2, 3, 4, 5, 6])
        for pattern in identity_patterns:
            assert numpy.array_equal(x, rearrange(x, pattern)), pattern

        for pattern1, pattern2 in equivalent_rearrange_patterns:
            assert numpy.array_equal(rearrange(x, pattern1), rearrange(x, pattern2))

        for reduction in ['min', 'max', 'sum']:
            for pattern1, pattern2 in equivalent_reduction_patterns:
                assert numpy.array_equal(reduce(x, pattern1, reduction=reduction),
                                         reduce(x, pattern2, reduction=reduction))

        # now just check coincidence with numpy
        all_rearrange_patterns = [*identity_patterns]
        for pattern_pairs in equivalent_rearrange_patterns:
            all_rearrange_patterns.extend(pattern_pairs)

    def test_collapsed_ellipsis_errors_out(self):
        x = numpy.zeros([1, 1, 1, 1, 1])
        rearrange(x, 'a b c d ... ->  a b c ... d')
        error = 0
        try:
            rearrange(x, 'a b c d (...) ->  a b c ... d')
        except Exception as e:
            error = 1
        assert error == 1

        rearrange(x, '... ->  (...)')
        error = 0
        try:
            rearrange(x, '(...) -> (...)')
        except Exception as e:
            error = 1
        assert error == 1

    def test_rearrange_imperative(self):
        for backend in collect_test_backends(symbolic=False, layers=True):
            print('Test layer for ', backend.framework_name)

            for pattern, axes_lengths, input_shape, wrong_shapes in rearrangement_patterns:
                x = numpy.arange(numpy.prod(input_shape), dtype='float32').reshape(input_shape)
                result_numpy = rearrange(x, pattern, **axes_lengths)
                layer = backend.layers().Rearrange(pattern, **axes_lengths)
                for shape in wrong_shapes:
                    try:
                        layer(backend.from_numpy(numpy.zeros(shape, dtype='float32')))
                    except:
                        pass
                    else:
                        raise AssertionError('Failure expected')

                # simple pickling / unpickling
                layer2 = pickle.loads(pickle.dumps(layer))
                result1 = backend.to_numpy(layer(backend.from_numpy(x)))
                result2 = backend.to_numpy(layer2(backend.from_numpy(x)))
                assert numpy.allclose(result_numpy, result1)
                assert numpy.allclose(result1, result2)

                just_sum = backend.layers().Reduce('...->', reduction='sum')

               
                variable = backend.from_numpy(x)
                result = just_sum(layer(variable))

                if 'jittor' in backend.framework_name:
                    grad = backend.jittor.grad(result, variable)
                else:
                    result.backward()
                    grad = variable.grad

                assert numpy.allclose(backend.to_numpy(grad), 1)

    def test_reduce_imperative(self):
        for backend in collect_test_backends(symbolic=False, layers=True):
            print('Test layer for ', backend.framework_name)
            for reduction in _reductions:
                for pattern, axes_lengths, input_shape, wrong_shapes in reduction_patterns:
                    print(backend, reduction, pattern, axes_lengths, input_shape, wrong_shapes)
                    x = numpy.arange(1, 1 + numpy.prod(input_shape), dtype='float32').reshape(input_shape)
                    x /= x.mean()
                    result_numpy = reduce(x, pattern, reduction, **axes_lengths)
                    layer = backend.layers().Reduce(pattern, reduction, **axes_lengths)
                    for shape in wrong_shapes:
                        try:
                            layer(backend.from_numpy(numpy.zeros(shape, dtype='float32')))
                        except:
                            pass
                        else:
                            raise AssertionError('Failure expected')

                    # simple pickling / unpickling
                    layer2 = pickle.loads(pickle.dumps(layer))
                    result1 = backend.to_numpy(layer(backend.from_numpy(x)))
                    result2 = backend.to_numpy(layer2(backend.from_numpy(x)))
                    assert numpy.allclose(result_numpy, result1)
                    assert numpy.allclose(result1, result2)

                    just_sum = backend.layers().Reduce('...->', reduction='sum')

                
                    variable = backend.from_numpy(x)
                    result = just_sum(layer(variable))

                    if 'jittor' in backend.framework_name:
                        grad = backend.jittor.grad(result, variable)
                        grad = backend.to_numpy(grad)
                    else:
                        result.backward()
                        grad = backend.to_numpy(variable.grad)
                    if reduction == 'sum':
                        assert numpy.allclose(grad, 1)
                    if reduction == 'mean':
                        assert numpy.allclose(grad, grad.min())
                    if reduction in ['max', 'min']:
                        assert numpy.all(numpy.in1d(grad, [0, 1]))
                        assert numpy.sum(grad) > 0.5

    def test_jittor_layer(self):
        has_jittor = any(backend.framework_name == 'jittor' for backend in collect_test_backends(symbolic=False, layers=True))
        if has_jittor:
            # checked that jittor present
            import jittor

            rtol = 1e-05
            atol = 1e-08
            def allclose(input, other): return jittor.all(jittor.abs(input-other) <= atol+rtol*jittor.abs(other))
            model1 = create_jittor_model(use_reduce=True)
            model2 = create_jittor_model(use_reduce=False)
            input = jittor.randn([10, 3, 32, 32])
            # random models have different predictions
            assert not allclose(model1(input), model2(input))
            model2.load_state_dict(pickle.loads(pickle.dumps(model1.state_dict())))
            assert allclose(model1(input), model2(input))


testcase = namedtuple('testcase', ['pattern', 'axes_lengths', 'input_shape', 'wrong_shapes'])

rearrangement_patterns = [
    testcase('b c h w -> b (c h w)', dict(c=20), (10, 20, 30, 40),
             [(), (10,), (10, 10, 10), (10, 21, 30, 40), [1, 20, 1, 1, 1]]),
    testcase('b c (h1 h2) (w1 w2) -> b (c h2 w2) h1 w1', dict(h2=2, w2=2), (10, 20, 30, 40),
             [(), (1, 1, 1, 1), (1, 10, 3), ()]),
    testcase('b ... c -> c b ...', dict(b=10), (10, 20, 30),
             [(), (10,), (5, 10)]),
]

reduction_patterns = rearrangement_patterns + [
    testcase('b c h w -> b ()', dict(b=10), (10, 20, 30, 40),
             [(10,), (10, 20, 30)]),
    testcase('b c (h1 h2) (w1 w2) -> b c h1 w1', dict(h1=15, h2=2, w2=2), (10, 20, 30, 40),
             [(10, 20, 31, 40)]),
    testcase('b ... c -> b', dict(b=10), (10, 20, 30, 40),
             [(10,), (11, 10)]),
]

equivalent_reduction_patterns = [
    ('a b c d e -> ', ' ... ->  '),
    ('a b c d e -> (e a)', 'a ... e -> (e a)'),
    ('a b c d e -> d (a e)', ' a b c d e ... -> d (a e) '),
    ('a b c d e -> (a b)', ' ... c d e  -> (...) '),
]

equivalent_rearrange_patterns = [
    ('a b c d e -> (a b) c d e', 'a b ... -> (a b) ... '),
    ('a b c d e -> a b (c d) e', '... c d e -> ... (c d) e'),
    ('a b c d e -> a b c d e', '... -> ... '),
    ('a b c d e -> (a b c d e)', '... ->  (...)'),
    ('a b c d e -> b (c d e) a', 'a b ... -> b (...) a'),
    ('a b c d e -> b (a c d) e', 'a b ... e -> b (a ...) e'),
]

identity_patterns = [
    '...->...',
    'a b c d e-> a b c d e',
    'a b c d e ...-> ... a b c d e',
    'a b c d e ...-> a ... b c d e',
    '... a b c d e -> ... a b c d e',
    'a ... e-> a ... e',
    'a ... -> a ... ',
    'a ... c d e -> a (...) c d e',
]

test_cases_repeat_anonymous = [
    # all assume that input has shape [1, 2, 4, 6]
    ('a b c d -> c a d b', dict()),
    ('a b c d -> (c 2 d a b)', dict(a=1, c=4, d=6)),
    ('1 b c d -> (d copy 1) 3 b c ', dict(copy=3)),
    ('1 ...  -> 3 ... ', dict()),
    ('() ... d -> 1 (copy1 d copy2) ... ', dict(copy1=2, copy2=3)),
    ('1 b c d -> (1 1) (1 b) 2 c 3 d (1 1)', dict()),

]

repeat_test_cases = [
    # all assume that input has shape [2, 3, 5]
    ('a b c -> c a b', dict()),
    ('a b c -> (c copy a b)', dict(copy=2, a=2, b=3, c=5)),
    ('a b c -> (a copy) b c ', dict(copy=1)),
    ('a b c -> (c a) (copy1 b copy2)', dict(a=2, copy1=1, copy2=2)),
    ('a ...  -> a ... copy', dict(copy=4)),
    ('... c -> ... (copy1 c copy2)', dict(copy1=1, copy2=2)),
    ('...  -> ... ', dict()),
    (' ...  -> copy1 ... copy2 ', dict(copy1=2, copy2=3)),
    ('a b c  -> copy1 a copy2 b c () ', dict(copy1=2, copy2=1)),
]


def check_reversion(x, repeat_pattern, **sizes):
    """Checks repeat pattern by running reduction """
    left, right = repeat_pattern.split('->')
    reduce_pattern = right + '->' + left
    repeated = reduce(x, repeat_pattern, reduction='repeat', **sizes)
    reduced_min = reduce(repeated, reduce_pattern, reduction='min', **sizes)
    reduced_max = reduce(repeated, reduce_pattern, reduction='max', **sizes)
    assert numpy.array_equal(x, reduced_min)
    assert numpy.array_equal(x, reduced_max)


def check_op_against_numpy(backend, numpy_input, pattern, axes_lengths, reduction='rearrange', is_symbolic=False):
    """
    Helper to test result of operation (rearrange or transpose) against numpy
    if reduction == 'rearrange', rearrange op is tested, otherwise reduce
    """
    if len(numpy_input.shape) == 0:
        return

    def operation(x):
        if reduction == 'rearrange':
            return rearrange(x, pattern, **axes_lengths)
        else:
            return reduce(x, pattern, reduction, **axes_lengths)

    numpy_result = operation(numpy_input)
    check_equal = numpy.array_equal
    p_none_dimension = 0.5
    if 'jittor' in backend.framework_name:
        check_equal = numpy.allclose
        p_none_dimension = 0
    if is_symbolic:
        symbol_shape = [d if numpy.random.random() >= p_none_dimension else None for d in numpy_input.shape]
        symbol = backend.create_symbol(shape=symbol_shape)
        result_symbol = operation(symbol)
        backend_result = backend.eval_symbol(result_symbol, [(symbol, numpy_input)])
    else:
        backend_result = operation(backend.from_numpy(numpy_input))
        backend_result = backend.to_numpy(backend_result)

    check_equal(numpy_result, backend_result)


def create_jittor_model(use_reduce=False):
    from jittor.nn import Sequential, Conv2d, MaxPool2d, Linear, ReLU
    from jittor.einops.layers.jittor import Rearrange, Reduce, EinMix
    return Sequential(
        Conv2d(3, 6, kernel_size=(5, 5)),
        Reduce('b c (h h2) (w w2) -> b c h w', 'max', h2=2, w2=2) if use_reduce else MaxPool2d(kernel_size=2),
        Conv2d(6, 16, kernel_size=(5, 5)),
        Reduce('b c (h h2) (w w2) -> b c h w', 'max', h2=2, w2=2),
        Rearrange('b c h w -> b (c h w)'),
        Linear(16 * 5 * 5, 120),
        ReLU(),
        Linear(120, 84),
        ReLU(),
        EinMix('b c1 -> (b c2)', weight_shape='c1 c2', bias_shape='c2', c1=84, c2=84),
        EinMix('(b c2) -> b c3', weight_shape='c2 c3', bias_shape='c3', c2=84, c3=84),
        Linear(84, 10),
    )


if __name__ == '__main__':
    unittest.main()


# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers:
#     Guoye Yang <498731903@qq.com>
#     Dun Liang <randonlang@gmail.com>.
#
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************

'''
example:

export class_home=/mnt/disk/cjld/class_nn
mkdir -p $class_home
docker pull jittor/jittor-cuda
python3.7 -m jittor_utils.class.setup_env setup 4
python3.7 -m jittor_utils.class.setup_env start 4
python3.7 -m jittor_utils.class.setup_env report
python3.7 -m jittor_utils.class.setup_env restart 4
python3.7 -m jittor_utils.class.setup_env stop
'''
# export class_home
# setup [n]          // setup for n users. including build user paths, user_info.txt and docker imgs. !!!WILL RESET SUDENT_FILES!!!
# start [n_gpu]      // run n docker CONTAINERs with n_gpu GPUs.
# stop               // stop n docker CONTAINERs
# restart [n_gpu]    // restart n docker CONTAINERs with n_gpu GPUs.
import sys
import os 
import json as js
import random

class_home = os.environ["class_home"]
student_files_dir = class_home + "/student_files"
student_files_bk_dir = class_home + "/student_files_bak"
cwd = os.path.dirname(__file__)

def run_cmd(cmd):
    print("[CMD]:", cmd)
    ret = os.system(cmd)
    if ret:
        print("[CMD] return", ret)
    return ret

def generate_random_str(randomlength):
  random_str = ''
  base_str = 'ABCDEFGHIGKLMNOPQRSTUVWXYZabcdefghigklmnopqrstuvwxyz0123456789'
  length = len(base_str) - 1
  for i in range(randomlength):
    random_str += base_str[random.randint(0, length)]
  return random_str

def setup(n):
    if os.path.exists(student_files_dir):
        if os.path.exists(student_files_bk_dir):
            run_cmd(f"rm -rf {student_files_bk_dir}")
        run_cmd(f"mv {student_files_dir} {student_files_bk_dir}")
    os.makedirs(student_files_dir)
    user_info = []
    for i in range(n): # 0 for root
        port = 20000 + i
        passwd = generate_random_str(8)
        name = 'stu_'+str(i)
        path = os.path.abspath(os.path.join(student_files_dir, name))
        info = {'port': port,
                'passwd': passwd,
                'name': name,
                'path': path}
        user_info.append(info)
        student_files_src = class_home + "/student_files_src"
        if os.path.isdir(student_files_src):
            run_cmd(f"cp -r {student_files_src} {path}")
        else:
            run_cmd('mkdir -p ' + path)
    js.dump(user_info, open(student_files_dir+"/user_info.json", "w"))

def start(n, n_gpu):
    assert os.path.exists(student_files_dir+'/user_info.json')
    user_info = js.load(open(student_files_dir+'/user_info.json', 'r'))
    for i in range(len(user_info)):
        id = i % n
        ids = ''
        for j in range(n_gpu):
            if j > 0:
                ids+=','
            ids += str((i * n_gpu + j) % n)
        u = user_info[i]
        print('START', i, '/', len(user_info))
        assert 0 == run_cmd(f'docker run -itd --shm-size=8g --network host --name {u["name"]} -v {u["path"]}:/root --gpus \'"device={ids}"\' jittor/jittor-cuda bash')
        # assert 0 == run_cmd(f'docker exec -it {u["name"]} bash -c \'apt update && apt install openssh-server -y\'')
        assert 0 == run_cmd(f'docker cp {cwd}/setup.py {u["name"]}:/etc/ssh/setup.py')
        assert 0 == run_cmd(f'docker cp {cwd}/motd {u["name"]}:/etc/motd')
        assert 0 == run_cmd(f'docker exec -it {u["name"]} python3.7 /etc/ssh/setup.py passwd {u["passwd"]}')
        assert 0 == run_cmd(f'docker exec -it {u["name"]} python3.7 /etc/ssh/setup.py ssh {u["port"]}')
        assert 0 == run_cmd(f'docker exec -it {u["name"]} python3.7 -m pip install jittor -U')
        assert 0 == run_cmd(f'docker exec -it {u["name"]} python3.7 -m jittor.test.test_example')

def stop():
    assert os.path.exists(student_files_dir+'/user_info.json')
    user_info = js.load(open(student_files_dir+'/user_info.json', 'r'))
    for i in range(len(user_info)):
        u = user_info[i]
        print('STOP', i, '/', len(user_info))
        run_cmd(f'docker rm -f {u["name"]}')

def report():
    assert os.path.exists(student_files_dir+'/user_info.json')
    user_info = js.load(open(student_files_dir+'/user_info.json', 'r'))
    hostname = open("/etc/hostname", 'r').read().strip() + ".randonl.me"
    for i in range(len(user_info)):
        u = user_info[i]
        print(f"ssh -p {u['port']} root@{hostname} # passwd: {u['passwd']}")

def restart(n, n_gpu):
    stop()
    start(n, n_gpu)

args = sys.argv[1:]
if (args[0] == 'setup'):
    assert(len(args) == 2)
    assert(type(eval(args[1])) == int)
    n = int(args[1])
    assert(n < 999)
    setup(n)
elif (args[0] == 'start'):
    assert(len(args) in [2,3])
    assert(type(eval(args[1])) == int)
    n = int(args[1])
    if len(args) == 3:
        assert(type(eval(args[2])) == int)
        n_gpu = int(args[2])
    else:
        n_gpu=1
    start(n, n_gpu)
elif (args[0] == 'stop'):
    stop()
elif (args[0] == 'restart'):
    assert(len(args) in [2,3])
    assert(type(eval(args[1])) == int)
    n = int(args[1])
    if len(args) == 3:
        assert(type(eval(args[2])) == int)
        n_gpu = int(args[2])
    else:
        n_gpu=1
    restart(n, n_gpu)
elif (args[0] == 'report'):
    report()
else:
    assert(False)




# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers:
#     Guoye Yang <498731903@qq.com>
#     Dun Liang <randonlang@gmail.com>.
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
import jittor.models as jtmodels

def load_parameters(m1, m2):
    m1.save('/tmp/temp.pk')
    m2.load('/tmp/temp.pk')

def compare_parameters(m1, m2):
    ps1 = m1.parameters()
    ps2 = m2.parameters()
    for i in range(len(ps1)):
        x = ps1[i].data + 1e-8
        y = ps2[i].data + 1e-8
        relative_error = abs(x - y) / abs(y)
        diff = relative_error.mean()
        assert diff < 1e-4, (diff, 'backward', ps2[i].name(), ps1[i].mean(), ps1[i].std(), ps2[i].mean(), ps2[i].std())

class TestDepthwiseConv(unittest.TestCase):
    @unittest.skipIf(not jt.has_cuda, "Cuda not found")
    @jt.flag_scope(use_cuda=1)
    def test_data(self):
        test_img = np.random.random((64,3,224,224)).astype('float32')
        jittor_test_img = jt.array(test_img)
        lr = 100

        jittor_model = jtmodels.__dict__['mobilenet_v2']()
        jittor_model2 = jtmodels.__dict__['mobilenet_v2']()
        # Set eval to avoid dropout layer & bn errors
        jittor_model.train()
        jittor_model.classifier[0].eval()
        for m in jittor_model.modules():
            if isinstance(m, jt.nn.BatchNorm):
                m.eval()

        jittor_model2.train()
        jittor_model2.classifier[0].eval()
        for m in jittor_model2.modules():
            if isinstance(m, jt.nn.BatchNorm):
                m.eval()

        load_parameters(jittor_model2, jittor_model)
        for m in jittor_model.modules():
            if isinstance(m, jt.nn.Conv):
                m.is_depthwise_conv = False
        cnt = 0
        for m in jittor_model2.modules():
            if isinstance(m, jt.nn.Conv):
                if (m.is_depthwise_conv):
                    cnt += 1
        assert cnt == 17, (cnt, '!=', 17)
        jt_optimizer = jt.nn.SGD(jittor_model.parameters(), lr = lr)
        jt_optimizer2 = jt.nn.SGD(jittor_model2.parameters(), lr = lr)

        jittor_result = jittor_model(jittor_test_img)
        mask = jt.random(jittor_result.shape, jittor_result.dtype)
        loss = jittor_result * mask
        jt_optimizer.step(loss)
        jt.sync_all(True)

        jittor_result2 = jittor_model2(jittor_test_img)
        loss = jittor_result2 * mask

        x = jittor_result2.data + 1e-8
        y = jittor_result.data + 1e-8
        relative_error = abs(x - y) / abs(y)
        diff = relative_error.mean()
        assert diff < 1e-4, (diff, 'forword')

        jt_optimizer2.step(loss)
        jt.sync_all(True)
        compare_parameters(jittor_model, jittor_model2)


        jt.clean()
        jt.gc()

if __name__ == "__main__":
    unittest.main()


import jittor as jt
import os
import jittor_utils
from jittor_utils import lock
import jittor.compiler as compiler
import numpy as np

dirname = os.path.dirname(__file__)
cc_files = [ dirname + "/vcompiler.cc"]
with open(dirname + "/vcompiler.h") as f:
    h_src = f.read()

with lock.lock_scope():
    mod = jittor_utils.compile_module(h_src, compiler.cc_flags + "-I" + dirname + " " + " ".join(cc_files))

for k, v in mod.__dict__.items():
    if k.startswith("_"):
        continue
    globals()[k] = v

def dfs(obj, path=""):
    if isinstance(obj, jt.Var):
        return [((path, len(obj.shape), str(obj.dtype)), obj)]
    if isinstance(obj, (list, tuple)):
        ret = []
        for i, v in enumerate(obj):
            ret += dfs(v, path + "[%d]" % i)
        return ret
    if isinstance(obj, dict):
        ret = []
        for k, v in obj.items():
            ret += dfs(v, path + "[%r]" % k)
        return ret
    return []

def dfs_config(obj):
    if isinstance(obj, jt.Var):
        return "Var"
    if isinstance(obj, (int, float, bool, str, type(None))):
        return obj
    if isinstance(obj, (list, tuple)):
        return [ dfs_config(v) for v in obj ]
    if isinstance(obj, dict):
        return { k:dfs_config(v) for k, v in obj.items() }
    raise ValueError(f"Unknown type {type(obj)}")

def dfs_clone_var(obj):
    if isinstance(obj, jt.Var):
        return obj.clone()
    if isinstance(obj, (int, float, bool, str, type(None))):
        return obj
    if isinstance(obj, (list, tuple)):
        return [ dfs_clone_var(v) for v in obj ]
    if isinstance(obj, dict):
        return { k:dfs_clone_var(v) for k, v in obj.items() }
    raise ValueError(f"Unknown type {type(obj)}")

def dfs_fill(obj, vars):
    i = 0
    def dfs_fill_var(obj):
        nonlocal i
        if isinstance(obj, jt.Var):
            v = vars[i]
            i += 1
            return v
        if isinstance(obj, (int, float, bool, str, type(None))):
            return obj
        if isinstance(obj, (list, tuple)):
            return [ dfs_fill_var(v) for v in obj ]
        if isinstance(obj, dict):
            ret = { k:dfs_fill_var(v) for k, v in obj.items() }
            return obj.__class__(ret)
        raise ValueError(f"Unknown type {type(obj)}")
    return dfs_fill_var(obj)


class CachedGraph:
    def __init__(self, func, args, kw):
        args = dfs_clone_var(args)
        kw = dfs_clone_var(kw)
        self.func = func
        self.inputs = (args, kw)
        jt.sync_all()
        exec_called = jt.flags.exec_called
        self.outputs = func(*args, **kw)
        import gc; gc.collect()
        assert exec_called == jt.flags.exec_called
        self.outputs_parsed = dfs(self.outputs)
        self.outputs_var = [ v for _, v in self.outputs_parsed ]
        self.inputs_parsed = dfs(self.inputs)
        self.inputs_var = [ v for _, v in self.inputs_parsed ]
        self.inputs_key = str([ key for key, _ in self.inputs_parsed ])
        for v in self.outputs_var:
            v.release_from_holders()
        for v in self.inputs_var:
            v.release_from_holders()
        self.sgraph = mod.build_sgraph(self.outputs_var, self.inputs_var)

# a function decorator
# build new graph:
# 1. shape dim changed
# 2. dtype changed
# 3. var path changed
# graph key:
# (args, kw), [ (var_path, shape dim, dtype), var ]
def build(func, debug=False):
    cache = {}
    def func_wrapper(*args, **kw):
        inputs = (args, kw)
        config_key = str(dfs_config(inputs))
        inputs_parsed = dfs(inputs)
        inputs_key = str([ key for key, _ in inputs_parsed ])
        inputs_var = [ v for _, v in inputs_parsed ]
        jt.sync(inputs_var)
        all_key = config_key + inputs_key
        if all_key not in cache:
            print(f"create graph with key '{all_key[:30]}'...")
            cache[all_key] = CachedGraph(func, args, kw)
        graph = cache[all_key]
        if not mod.prob_sgraph(graph.sgraph, inputs_var):
            print(f"merge graph with key '{all_key[:30]}'...")
            graph2 = CachedGraph(func, args, kw)
            mod.merge_sgraph(graph.sgraph, graph2.sgraph)
        outputs = mod.exec_sgraph(graph.sgraph, inputs_var)
        if debug:
            graph2 = CachedGraph(func, args, kw)
            outputs2 = mod.exec_sgraph(graph2.sgraph, inputs_var)
            for v1, v2 in zip(outputs, outputs2):
                np.testing.assert_allclose(v1.data, v2.data, rtol=0.01, atol=0.05)
        return dfs_fill(graph.outputs, outputs)
    return func_wrapper

# c interface
# build_sgraph -> sgraph
# merge_sgraph
# exec_sgraph
# prob_sgraph: check sgraph can exec
# if prob_sgraph failed
#   build_sgraph
#   merge_sgraph
# exec_sgraph



# overall code:
# 1. get input_key from (args, kw)
# 2. if input_key not in cache
#       graph.outputs = func(*args, **kw)
#       graph.outputs_var = var_parser(graph.outputs)
#       graph.sgraph = build_sgraph(outputs_var)
#       graph.inputs = (args, kw)
#       graph.inputs_var = var_parser(args, kw)



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import os
import subprocess as sp
import sys

class TestTracer(unittest.TestCase):
    def test_print_trace(self):
        jt.print_trace()

        if os.name != 'nt':
            # force use addr2line
            with jt.flag_scope(gdb_path=""):
                jt.print_trace()

    def test_breakpoint(self):
        fname = os.path.join(jt.flags.cache_path, "test_breakpoint.py")
        with open(fname, 'w') as f:
            f.write("""
import jittor as jt
with jt.flag_scope(extra_gdb_cmd="c;q"):
    jt.flags.gdb_attach = 1
""")
        out = sp.getoutput(sys.executable+' '+fname)
        print(out)
        assert "Attaching to" in out

    def test_segfault(self):
        if os.name == 'nt':
            a = jt.array([1,2,3])
            b = jt.array([1,2,300000000])
            c = a[b]
            try:
                c.sync()
            except Exception as e:
                assert "access violation reading" in str(e)
        


if __name__ == "__main__":
    unittest.main()


import os

import numpy as np
import jittor as jt
from jittor import nn

f = open(os.path.join(os.path.realpath(os.path.dirname(__file__)), "src", "igamma.h"), "r")
cuda_header = f.read()
f.close()

def igamma(alpha, x):
    cuda_src = '''
        @alias(x, in0)
        @alias(px ,out0)
        int batch_size = x_stride0 == 1 ? 1 : x_shape0;
        int batch_shape = x_shape0 * x_stride0 / batch_size;
        float alpha = data["alpha"];
        igamma_kernel<<<batch_size, 16>>>(x_p, px_p, alpha, batch_shape);   
    '''
    out = jt.code(x.shape, x.dtype, [x], cuda_header=cuda_header, cuda_src=cuda_src, data={"alpha": alpha})
    return out



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import os
import numpy as np
import jittor.nn as jnn

skip_this_test = False

try:
    jt.dirty_fix_pytorch_runtime_error()
    import torch
    import torch.nn as tnn
except:
    skip_this_test = True

@unittest.skipIf(skip_this_test, "No Torch found")
class TestLoss(unittest.TestCase):
    def test_l1_loss(self):
        jt_loss=jnn.L1Loss()
        tc_loss=tnn.L1Loss()
        output=np.random.randn(10,100).astype(np.float32)
        target=np.random.randn(10,100).astype(np.float32)
        jt_y=jt_loss(jt.array(output), jt.array(target))
        tc_y=tc_loss(torch.from_numpy(output), torch.from_numpy(target))
        assert np.allclose(jt_y.numpy(), tc_y.numpy())
        
    def test_mse_loss(self):
        jt_loss=jnn.MSELoss()
        tc_loss=tnn.MSELoss()
        output=np.random.randn(10,100).astype(np.float32)
        target=np.random.randn(10,100).astype(np.float32)
        jt_y=jt_loss(jt.array(output), jt.array(target))
        tc_y=tc_loss(torch.from_numpy(output), torch.from_numpy(target))
        assert np.allclose(jt_y.numpy(), tc_y.numpy())
        
    def test_nll_loss(self):
        tc_loss = tnn.functional.nll_loss
        jt_loss = jnn.nll_loss
        output=np.random.randn(10,10).astype(np.float32)
        target=np.random.randint(10, size=(10))
        jt_y=jt_loss(jt.array(output), jt.array(target),reduction='mean')
        tc_y=tc_loss(torch.from_numpy(output), torch.from_numpy(target),reduction='mean')
        assert np.allclose(jt_y.numpy(), tc_y.numpy())
        output=np.random.randn(10,10).astype(np.float32)
        target=np.random.randint(10, size=(10))
        weight=np.random.randn(10,).astype(np.float32)
        jt_y=jt_loss(jt.array(output), jt.array(target),jt.array(weight),reduction='mean')
        tc_y=tc_loss(torch.from_numpy(output), torch.from_numpy(target),torch.from_numpy(weight),reduction='mean')
        assert np.allclose(jt_y.numpy(), tc_y.numpy())

    def test_cross_entropy_loss(self):
        jt_loss=jnn.CrossEntropyLoss()
        tc_loss=tnn.CrossEntropyLoss()
        output=np.random.randn(10,10).astype(np.float32)
        target=np.random.randint(10, size=(10))
        jt_y=jt_loss(jt.array(output), jt.array(target))
        tc_y=tc_loss(torch.from_numpy(output), torch.from_numpy(target))
        assert np.allclose(jt_y.numpy(), tc_y.numpy())
    
    def test_cross_entropy_loss_v2(self):
        B = 100
        C = 5
        for shape in [[100,1],[],[100,20]]:
            s1 = [B,C]+shape
            s2 = [B]+shape
            a = np.random.randn(*s1).astype(np.float32)
            b = np.random.randint(0,C,size=s2).astype(np.int32)
            weight = np.random.randn(C).astype(np.float32)

            for r in ['mean','sum','none']:
                r1 = torch.nn.functional.cross_entropy(torch.tensor(a),torch.tensor(b.astype(np.int64)),weight=torch.tensor(weight),reduction=r)
                r2 = jnn.cross_entropy_loss(jt.array(a),jt.array(b),weight=jt.array(weight),reduction=r)
                np.testing.assert_allclose(r1.numpy(),r2.numpy(),rtol=1e-3, atol=1e-3)
            
            for r in ['mean','sum','none']:
                r1 = torch.nn.functional.cross_entropy(torch.tensor(a),torch.tensor(b.astype(np.int64)),reduction=r)
                r2 = jnn.cross_entropy_loss(jt.array(a),jt.array(b),reduction=r)
                np.testing.assert_allclose(r1.numpy(),r2.numpy(),rtol=1e-3, atol=1e-3)
            
            r1 = torch.nn.functional.cross_entropy(torch.tensor(a),torch.tensor(b.astype(np.int64)))
            r2 = jnn.cross_entropy_loss(jt.array(a),jt.array(b))
            np.testing.assert_allclose(r1.numpy(),r2.numpy(),rtol=1e-3, atol=1e-3)

            r1 = torch.nn.functional.cross_entropy(torch.tensor(a),torch.tensor(b.astype(np.int64)),weight=torch.tensor(weight))
            r2 = jnn.cross_entropy_loss(jt.array(a),jt.array(b),weight=jt.array(weight))
            np.testing.assert_allclose(r1.numpy(),r2.numpy(),rtol=1e-3, atol=1e-3)

            for r in ['mean','sum','none']:
                r1 = torch.nn.functional.cross_entropy(torch.tensor(a),torch.tensor(b.astype(np.int64)),weight=torch.tensor(weight),reduction=r,ignore_index=C//2)
                r2 = jnn.cross_entropy_loss(jt.array(a),jt.array(b),weight=jt.array(weight),reduction=r,ignore_index=C//2)
                np.testing.assert_allclose(r1.numpy(),r2.numpy(),rtol=1e-3, atol=1e-3)
            
            for r in ['mean','sum','none']:
                r1 = torch.nn.functional.cross_entropy(torch.tensor(a),torch.tensor(b.astype(np.int64)),reduction=r,ignore_index=C//2)
                r2 = jnn.cross_entropy_loss(jt.array(a),jt.array(b),reduction=r,ignore_index=C//2)
                np.testing.assert_allclose(r1.numpy(),r2.numpy(),rtol=1e-3, atol=1e-3)
            
            r1 = torch.nn.functional.cross_entropy(torch.tensor(a),torch.tensor(b.astype(np.int64)),ignore_index=C//2)
            r2 = jnn.cross_entropy_loss(jt.array(a),jt.array(b),ignore_index=C//2)
            np.testing.assert_allclose(r1.numpy(),r2.numpy(),rtol=1e-3, atol=1e-3)

            r1 = torch.nn.functional.cross_entropy(torch.tensor(a),torch.tensor(b.astype(np.int64)),weight=torch.tensor(weight),ignore_index=C//2)
            r2 = jnn.cross_entropy_loss(jt.array(a),jt.array(b),weight=jt.array(weight),ignore_index=C//2)
            np.testing.assert_allclose(r1.numpy(),r2.numpy(),rtol=1e-3, atol=1e-3)


    def test_cross_entropy_ignore_index(self):
        ignore_index = np.random.randint(0, 10)
        jt_loss = jnn.CrossEntropyLoss(ignore_index=ignore_index)
        tc_loss = tnn.CrossEntropyLoss(ignore_index=ignore_index)
        output = np.random.rand(100, 10).astype(np.float32)
        target = np.random.randint(10, size=(100))
        jt_y=jt_loss(jt.array(output), jt.array(target))
        tc_y=tc_loss(torch.from_numpy(output), torch.from_numpy(target))
        assert np.allclose(jt_y.numpy(), tc_y.numpy())

    def test_cross_entropy_weight(self):
        weight = np.random.rand(10).astype('float32')
        jt_loss = jnn.CrossEntropyLoss(weight=jt.array(weight))
        tc_loss = tnn.CrossEntropyLoss(weight=torch.from_numpy(weight))
        output = np.random.rand(100, 10).astype(np.float32)
        target = np.random.randint(10, size=(100))
        jt_y=jt_loss(jt.array(output), jt.array(target))
        tc_y=tc_loss(torch.from_numpy(output), torch.from_numpy(target))
        assert np.allclose(jt_y.numpy(), tc_y.numpy())

    def test_cross_entropy_weight_ignore(self):
        weight = np.random.rand(4).astype('float32')
        jt_loss = jnn.CrossEntropyLoss(weight=jt.array(weight), ignore_index=1)
        tc_loss = tnn.CrossEntropyLoss(weight=torch.from_numpy(weight), ignore_index=1)
        output = np.random.rand(3, 4, 2,2).astype(np.float32)
        target = np.random.randint(4, size=(3, 2,2))
        jt_y=jt_loss(jt.array(output), jt.array(target))
        tc_y=tc_loss(torch.from_numpy(output), torch.from_numpy(target))
        np.testing.assert_allclose(jt_y.numpy(), tc_y.numpy())


    def test_bce_loss(self):
        jt_loss=jnn.BCELoss()
        tc_loss=tnn.BCELoss()
        jt_sig = jnn.Sigmoid()
        tc_sig = tnn.Sigmoid()
        output=np.random.randn(100).astype(np.float32)
        target=np.random.randint(2, size=(100)).astype(np.float32)
        jt_y=jt_loss(jt_sig(jt.array(output)), jt.array(target))
        tc_y=tc_loss(tc_sig(torch.from_numpy(output)), torch.from_numpy(target))
        assert np.allclose(jt_y.numpy(), tc_y.numpy())

        weight=np.random.randn(100).astype(np.float32)
        jt_loss=jnn.BCELoss(weight=jt.array(weight), size_average=False)
        tc_loss=tnn.BCELoss(weight=torch.Tensor(weight), size_average=False)
        jt_y=jt_loss(jt_sig(jt.array(output)), jt.array(target))
        tc_y=tc_loss(tc_sig(torch.from_numpy(output)), torch.from_numpy(target))
        assert np.allclose(jt_y.numpy(), tc_y.numpy())
        
    def test_bce_with_logits_loss(self):
        jt_loss=jnn.BCEWithLogitsLoss()
        tc_loss=tnn.BCEWithLogitsLoss()
        output=np.random.randn(100).astype(np.float32)
        target=np.random.randint(2, size=(100)).astype(np.float32)
        jt_y=jt_loss(jt.array(output), jt.array(target))
        tc_y=tc_loss(torch.from_numpy(output), torch.from_numpy(target))
        assert np.allclose(jt_y.numpy(), tc_y.numpy())
        
if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import os, sys, shutil
import glob
import jittor_utils as jit_utils

cache_path = os.path.join(jit_utils.home(), ".cache", "jittor")

def callback(func, path, exc_info):
    print(f"remove \"{path}\" failed.")

def rmtree(path):
    if os.path.isdir(path):
        print(f"remove \"{path}\" recursive.")
        shutil.rmtree(path, onerror=callback)

def clean_all():
    rmtree(cache_path)

def clean_core():
    rmtree(cache_path+"/default")
    rmtree(cache_path+"/master")
    fs = glob.glob(cache_path+"/jt*")
    for f in fs: rmtree(f)

def clean_cuda():
    rmtree(cache_path+"/jtcuda")
    rmtree(cache_path+"/cutt")
    rmtree(cache_path+"/cub")
    rmtree(cache_path+"/nccl")

def clean_dataset():
    rmtree(cache_path+"/dataset")

def clean_swap():
    rmtree(cache_path+"/tmp")

def print_help():
    msg = "|".join(keys)
    print(f"Usage: {sys.executable} -m jittor_utils.clean_cache [{msg}]")
    exit()


keys = [ k[6:] for k in globals() if k.startswith("clean_") ]

if __name__ == "__main__":
    if len(sys.argv)==1:
        print_help()
    else:
        for k in sys.argv[1:]:
            if k not in keys:
                print_help()
            func = globals()["clean_"+k]
            func()



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
import jittor.nn as jnn

skip_this_test = False

try:
    jt.dirty_fix_pytorch_runtime_error()
    import torch
    import torch.nn as tnn
except:
    torch = None
    tnn = None
    skip_this_test = True

# TODO: more test
@unittest.skipIf(skip_this_test, "No Torch found")
class TestSearchSorted(unittest.TestCase):
    def test_origin(self):
        sorted = jt.array([[1, 3, 5, 7, 9], [2, 4, 6, 8, 10]])
        values = jt.array([[3, 6, 9], [3, 6, 9]])
        ret = jt.searchsorted(sorted, values)
        assert (ret == [[1, 3, 4], [1, 2, 4]]).all(), ret

        ret = jt.searchsorted(sorted, values, right=True)
        assert (ret == [[2, 3, 5], [1, 3, 4]]).all(), ret

        sorted_1d = jt.array([1, 3, 5, 7, 9])
        ret = jt.searchsorted(sorted_1d, values)
        assert (ret == [[1, 3, 4], [1, 3, 4]]).all(), ret

    @unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
    @jt.flag_scope(use_cuda=1)
    def test_cuda(self):
        self.test_origin()


    def test_searchsorted_cpu(self):
        for i in range(1,3):
            s = np.sort(np.random.rand(*((10,)*i)),-1)
            v = np.random.rand(*((10,)*i))
            s_jt = jt.array(s)
            v_jt = jt.array(v)
            s_tc = torch.from_numpy(s)
            v_tc = torch.from_numpy(v)

            y_tc = torch.searchsorted(s_tc, v_tc, right=True)
            y_jt = jt.searchsorted(s_jt, v_jt, right=True)
            assert np.allclose(y_jt.numpy(), y_tc.data)
            y_jt = jt.searchsorted(s_jt, v_jt, right=False)
            y_tc = torch.searchsorted(s_tc, v_tc, right=False)
            assert np.allclose(y_jt.numpy(), y_tc.data)

    @unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
    @jt.flag_scope(use_cuda=1)
    def test_searchsorted_gpu(self):
        self.test_searchsorted_cpu()

        

if __name__ == "__main__":
    unittest.main()


import jittor as jt

jt.flags.use_device = 1
n = 100000

with jt.profile_scope(10, 10) as rep:
    jt.code([2], "float32", [],
    cuda_header='''__global__ void kernel(float* a) {}''',
    cuda_src=f'''
    for (int i=0; i<{n}; i++) kernel<<<1,32>>>(out0_p);
    ''').sync()

avg_ns = float(rep[1][4]) / n
print("kernel launch overhead(ns):", avg_ns)



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#    Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np

class TestErrorMsg(unittest.TestCase):

    def test_error_msg(self):
        a = jt.array([3,2,1])
        b = jt.code(a.shape, a.dtype, [a],
            cpu_header="""
                #include <algorithm>
                @alias(a, in0)
                @alias(b, out)
            """,
            cpu_src="""
                for (int i=0; i<a_shape0; i++)
                    @b(i) = @a(i);
                std::sort(&@b(0), &@b(in0_shape0));
                throw std::runtime_error("???");
            """
        )
        msg = ""
        try:
            print(b)
        except Exception as e:
            msg = str(e)
        assert "[Reason]: ???" in msg
        assert "[Input]: int32[3,]" in msg
        assert "[OP TYPE]: code" in msg
        assert "[Async Backtrace]:" in msg

    @jt.flag_scope(trace_py_var=3)
    def test_error_msg_trace_py_var(self):
        a = jt.array([3,2,1])
        b = jt.code(a.shape, a.dtype, [a],
            cpu_header="""
                #include <algorithm>
                @alias(a, in0)
                @alias(b, out)
            """,
            cpu_src="""
                for (int i=0; i<a_shape0; i++)
                    @b(i) = @a(i);
                std::sort(&@b(0), &@b(in0_shape0));
                throw std::runtime_error("???");
            """
        )
        msg = ""
        try:
            print(b)
        except Exception as e:
            msg = str(e)
        print(msg)
        assert "[Reason]: ???" in msg
        assert "[Input]: int32[3,]" in msg
        assert "[OP TYPE]: code" in msg
        assert "[Async Backtrace]:" in msg
        assert "test_error_msg.py:" in msg



if __name__ == "__main__":
    unittest.main()